/* Here is some basic information about writing the EEPROM in WK2 */
/* Message commands are written exactly like they would have been entered 
 * in standalone mode. For example a speed change would be:

 /S15

 * / being a forward slash. The routine putEEPmsg translates the ASCII message into
 * Morse coding that WK2 understands. Likewise when reading a block back from WK2
 * it needs to be translated back into ASCII. getEEPmsg does this with help from morseToASCII.
 *
 ;-------------------------------------------
;	Embedded Message Command Defs
;-------------------------------------------
;
BCON_CMD  /B     ; Beacon time load
CALL_CMD  /C     ; Call message
DEC_CMD   /D     ; Decrement serial number
HSCW_CMD  /H     ; HSCW request
KEY_CMD	  /K     ; Key transmitter for 2 secs
SN_CMD    /N     ; Play serial number
PSE_CMD   /P	 ; Insert pause delay
QRSS_CMD  /Q     ; QRSS request
SPD_CMD   /S     ; Speed change
WAIT_CMD  /W     ; Delay for N sec
XSPD_CMD  /X     ; Cancel speed override
ANA0_CMD  /Y	 ; Analog channel 0
ANA1_CMD  /Z	 ; Analog channel 1
MSG1_CMD  /1     ; Jump to Message 1
MSG2_CMD  /2     ; Jump to Message 2
MSG3_CMD  /3     ; Jump to Message 3
MSG4_CMD  /4     ; Jump to Message 4
MSG5_CMD  /5     ; Jump to Message 5
MSG6_CMD  /6     ; Jump to Message 6
;

/*** Winkey EEPROM Block ***/

typedef struct {
    BYTE magic;     // 0x00
    BYTE modereg;   // 0x01
    BYTE oprrate;   // 0x02
    BYTE stconst;   // 0x03
    BYTE weight;    // 0x04
    BYTE lead_time; // 0x05
    BYTE tail_time; // 0x06
    BYTE minwpm;    // 0x07
    BYTE wpmrange;  // 0x08
    BYTE xtnd;      // 0x09
    BYTE kcomp;     // 0x0a
    BYTE farnswpm;  // 0x0b
    BYTE sampadj;   // 0x0c
    BYTE ratio;     // 0x0d
    BYTE pincfg;    // 0x0e
    BYTE k12mode;   // 0x0f
    BYTE cmdwpm;    // 0x10
    BYTE freeptr;   // 0x11
    BYTE msgptr1;   // 0x12
    BYTE msgptr2;   // 0x13
    BYTE msgptr3;   // 0x14
    BYTE msgptr4;   // 0x15
    BYTE msgptr5;   // 0x16
    BYTE msgptr6;   // 0x17
    BYTE msgs[MAXMSG]; // 0x18
}WKEEPROM, *LPWKEEPROM;

WKEEPROM wkState;

void wk_readEEP(void)
{
    BYTE buf[8];

    buf[0] = ADMIN_CMD;         // admin cmd
    buf[1] = ADMIN_DUMPEEP;     // dump EEP
    writeABuffer (buf, 2);
    readABuffer ((BYTE*)&wkState, sizeof(wkState));
}

void wk_writeEEP(void)
{
    BYTE buf[8];

    buf[0] = ADMIN_CMD;         // admin cmd
    buf[1] = ADMIN_LOADEEP;     // load EEP
    writeABuffer (buf, 2);
    writeABuffer ((BYTE*)&wkState, sizeof(wkState));
}


/* Use the following function to take information out of an EEPROM structure
 * and initialize a dialog box. Usually used to unpack an EEPROM block
 * just read from WK2.
 */

/****************************************************
 * Initalize dialog state
 ***************************************************/
void readEEPimage (HANDLE hwnd) {

    int select;
    int index, i;
    char sbuf[256];

    SetDlgItemInt (hwnd, IDC_WEIGHT, wkState.weight, FALSE);
    SetDlgItemInt (hwnd, IDC_LEADIN, wkState.lead_time, FALSE);
    SetDlgItemInt (hwnd, IDC_TAIL, wkState.tail_time, FALSE);
    SetDlgItemInt (hwnd, IDC_EXTEND, wkState.xtnd, FALSE);
    SetDlgItemInt (hwnd, IDC_KEYCOMP, wkState.kcomp, FALSE);
    SetDlgItemInt (hwnd, IDC_FARNS, wkState.farnswpm, FALSE);
    SetDlgItemInt (hwnd, IDC_KSAMP, wkState.sampadj, FALSE);
    SetDlgItemInt (hwnd, IDC_RATIO, wkState.ratio, FALSE);
    SetDlgItemInt (hwnd, IDC_SETWPM2, wkState.oprrate, FALSE);

    switch (wkState.modereg & (NIAMBIC+KEYERMODE)) {
    case 00:
        select=1;
        break;
    case KEYERMODE:
        select=0;
        break;
    case NIAMBIC:
        if (wkState.pincfg & DITPRI)
            select=3;
        else if (wkState.pincfg & DAHPRI)
            select=4;
        else
            select=2;
        break;
    case KEYERMODE+NIAMBIC:
        select=5;
        break;
    }
    SendDlgItemMessage (hwnd, IDC_MODECOMBO, CB_SETCURSEL, select, 0);

    switch (wkState.pincfg & (HANG1+HANG0)) {
    case 0:
        select=0;
        break;
    case HANG0:
        select=1;
        break;
    case HANG1:
        select=2;
        break;
    case HANG1+HANG0:
        select=3;
        break;
    }
    SendDlgItemMessage (hwnd, IDC_HANGCOMBO, CB_SETCURSEL, select, 0);

    switch (wkState.pincfg & (KEYPORT1+KEYPORT0+USESTONE)) {
    case KEYPORT0:
        select=0;
        break;
    case KEYPORT0+USESTONE:
        select=1;
        break;
    case KEYPORT1:
        select=2;
        break;
    case KEYPORT1+USESTONE:
        select=3;
        break;
    }
    SendDlgItemMessage (hwnd, IDC_OUTPCOMBO, CB_SETCURSEL, select, 0);

    SendDlgItemMessage (hwnd, IDC_STFREQCOMBO, CB_SETCURSEL, wkState.stconst-1, 0);

    SetDlgItemInt (hwnd, IDC_TOPWPM, wkState.minwpm+wkState.wpmrange, FALSE);
    SetDlgItemInt (hwnd, IDC_BOTWPM, wkState.minwpm, FALSE);
    SendMessage (GetDlgItem(hwnd, IDC_SWAP), BM_SETCHECK, wkState.modereg & PDL_SWAP, 0L);
    SendMessage (GetDlgItem(hwnd, IDC_AUTOSP), BM_SETCHECK, wkState.modereg & ASPACE, 0L);
    SendMessage (GetDlgItem(hwnd, IDC_CTSPACE), BM_SETCHECK, wkState.modereg & CT_SPACE, 0L);
    SendMessage (GetDlgItem(hwnd, IDC_PDLDOG), BM_SETCHECK, !(wkState.modereg & NONSTICK), 0L);
    SendMessage (GetDlgItem(hwnd, IDC_CUT_0_9), BM_SETCHECK, wkState.k12mode & CUT09, 0L);

    getEEPmsg(sbuf, wkState.msgptr1);
    SetDlgItemText(hDlg, IDC_EDITMSG1, sbuf);

    getEEPmsg(sbuf, wkState.msgptr2);
    SetDlgItemText(hDlg, IDC_EDITMSG2, sbuf);

    getEEPmsg(sbuf, wkState.msgptr3);
    SetDlgItemText(hDlg, IDC_EDITMSG3, sbuf);

    getEEPmsg(sbuf, wkState.msgptr4);
    SetDlgItemText(hDlg, IDC_EDITMSG4, sbuf);

    getEEPmsg(sbuf, wkState.msgptr5);
    SetDlgItemText(hDlg, IDC_EDITMSG5, sbuf);

    getEEPmsg(sbuf, wkState.msgptr6);
    SetDlgItemText(hDlg, IDC_EDITMSG6, sbuf);

}

void getEEPmsg (BYTE* sbuf, BYTE index)
{
    BYTE i=0;
    int done=0;

    if (index!=16) {
        index -= 0x18;
        do {
            sbuf[i] = wkState.msgs[index++];
            if (sbuf[i] & 0x80) {
                sbuf[i] = morseToAscii(sbuf[i] & 0x7f);
                i++;
                done = 1;
            }
            else {
                sbuf[i] = morseToAscii(sbuf[i]);
                i++;
            }
        } while (!done);
        sbuf[i] = 0;
    }
    else {
        sbuf[i]=0;
    }
}


/* Use the following function to take data out of a dialog box and pack it
 * into an EEPROM block in preparation for downloading it to WK2
 */

BOOL buildEEPimage(HANDLE hwnd)
{
    int temp, botTemp, topTemp;
    BYTE index, next_index, i;
    BYTE sbuf[256];


    temp = GetDlgItemInt(hDlg, IDC_WEIGHT, NULL, FALSE);
    if ((temp<10) || (temp>90)) {
        MessageBox (hwnd, TEXT("Weight must be between 10 and 90"), TEXT("Error"), MB_OK);
        return FALSE;
    }
    wkState.weight = (BYTE)temp;

    temp = GetDlgItemInt(hDlg, IDC_LEADIN, NULL, FALSE);
    if (temp>250) {
        MessageBox (hwnd, TEXT("Lead In must be 250 or less."), TEXT("Error"), MB_OK);
    }
    wkState.lead_time = (BYTE)temp;

    temp = GetDlgItemInt(hDlg, IDC_TAIL, NULL, FALSE);
    if (temp>250) {
        MessageBox (hwnd, TEXT("Tail must be 250 or less."), TEXT("Error"), MB_OK);
    }
    wkState.tail_time = (BYTE)temp;

    temp = GetDlgItemInt(hDlg, IDC_EXTEND, NULL, FALSE);
    if (temp>250) {
        MessageBox (hwnd, TEXT("Extend must be 250 or less."), TEXT("Error"), MB_OK);
    }
    wkState.xtnd = (BYTE)temp;

    temp = GetDlgItemInt(hDlg, IDC_KEYCOMP, NULL, FALSE);
    if (temp>250) {
        MessageBox (hwnd, TEXT("Key Comp must be 250 or less."), TEXT("Error"), MB_OK);
    }
    wkState.kcomp = (BYTE)temp;

    temp = GetDlgItemInt(hDlg, IDC_FARNS, NULL, FALSE);
    if ( ((temp>=10) && (temp<=99))||(temp==0) ) {
        wkState.farnswpm = (BYTE)temp;
    }
    else {
        MessageBox (hwnd, TEXT("Farns must be 0 or between 10 and 90"), TEXT("Error"), MB_OK);
        return FALSE;
    }

    temp = GetDlgItemInt(hDlg, IDC_KSAMP, NULL, FALSE);
    if ((temp<10) || (temp>90)) {
        MessageBox (hwnd, TEXT("Ksamp must be between 10 and 90"), TEXT("Error"), MB_OK);
        return FALSE;
    }
    wkState.sampadj = (BYTE)temp;

    temp = GetDlgItemInt(hDlg, IDC_RATIO, NULL, FALSE);
    if ((temp<33) || (temp>66)) {
        MessageBox (hwnd, TEXT("Ratio must be between 33 and 66"), TEXT("Error"), MB_OK);
        return FALSE;
    }
    wkState.ratio = (BYTE)temp;

    topTemp = GetDlgItemInt(hDlg, IDC_TOPWPM, NULL, FALSE);
    botTemp = GetDlgItemInt(hDlg, IDC_BOTWPM, NULL, FALSE);
    if (topTemp <= botTemp) {
        MessageBox (hwnd, TEXT("Pot Max must be greater than Pot Min"), TEXT("Error"), MB_OK);
        return FALSE;
    }
    if ((botTemp<5) || (botTemp>99)) {
        MessageBox (hwnd, TEXT("Pot Min must be between 5 and 99"), TEXT("Error"), MB_OK);
        return FALSE;
    }
    if ((topTemp<5) || (topTemp>99)) {
        MessageBox (hwnd, TEXT("Pot Max must be between 5 and 99"), TEXT("Error"), MB_OK);
        return FALSE;
    }
    wkState.minwpm = (BYTE)botTemp;
    wkState.wpmrange = (BYTE)(topTemp-botTemp);

    index = 0x0;

    GetDlgItemText(hDlg, IDC_EDITMSG1, sbuf, 250);
    next_index = putEEPmsg(sbuf, index);
    if (next_index == 0) {
        wkState.msgptr1 = 16;
    }
    else {
        wkState.msgptr1 = index + 0x18;
        index = next_index;
    }

    GetDlgItemText(hDlg, IDC_EDITMSG2, sbuf, 250);
    next_index = putEEPmsg(sbuf, index);
    if (next_index == 0) {
        wkState.msgptr2 = 16;
    }
    else {
        wkState.msgptr2 = index + 0x18;
        index = next_index ;
    }

    GetDlgItemText(hDlg, IDC_EDITMSG3, sbuf, 250);
    next_index = putEEPmsg(sbuf, index);
    if (next_index == 0) {
        wkState.msgptr3 = 16;
    }
    else {
        wkState.msgptr3 = index + 0x18;
        index = next_index;
    }

    GetDlgItemText(hDlg, IDC_EDITMSG4, sbuf, 250);
    next_index = putEEPmsg(sbuf, index);
    if (next_index == 0) {
        wkState.msgptr4 = 16;
    }
    else {
        wkState.msgptr4 = index + 0x18;
        index = next_index;
    }

    GetDlgItemText(hDlg, IDC_EDITMSG5, sbuf, 250);
    next_index = putEEPmsg(sbuf, index);
    if (next_index == 0) {
        wkState.msgptr5 = 16;
    }
    else {
        wkState.msgptr5 = index + 0x18;
        index = next_index;
    }

    GetDlgItemText(hDlg, IDC_EDITMSG6, sbuf, 250);
    next_index = putEEPmsg(sbuf, index);
    if (next_index == 0) {
        wkState.msgptr6 = 16;
    }
    else {
        wkState.msgptr6 = index + 0x18;
        index = next_index;
    }

    wkState.freeptr = index + 0x18;

    return TRUE;
}

BYTE putEEPmsg (BYTE* sbuf, BYTE index)
{
    int i, count;
    BYTE temp;

    count = strlen(sbuf);
    if (count == 0) {
        return 0;
    }
    else {
        for (i=0; i<count; i++) {
            temp = sbuf[i];
            temp = ASCII_TO_MORSE[sbuf[i]-0x20];
            wkState.msgs[index++] = temp;
        }
        wkState.msgs[index-1] |= 0x80;
    }
    return index;
}


BYTE morseToAscii (BYTE morse)
{
    BYTE i;

    for (i=0; i<96; i++) {
        if (ASCII_TO_MORSE[i] == morse) {
            return i+0x20;
        }
    }
    return 0;
}

