//
//  WinKeyerUsb.m
//  SkookumMorse
//
//  Created by Willard Myers on 2009.04.23.
//  Copyright 2009 K1GQ. All rights reserved.
//

@import ORSSerial;

#import "WinKeyerConstants.h"
#import "PreferencesGeneral.h"
#import "PreferencesMessages.h"
#import "PreferencesWinKeyer.h"
#import "SkookumMorseDelegate.h"
#import "WinKeyerUsb.h"

// MARK: - WinKeyer Morse Character Definitions -
#define Morse_WARN  0x40    // H 0100 0000 = DI-DI-DI-DI-DI-DIT
#define Morse_BEG   0x1A    // 0001 1010 = DI-DAH-DI-DAH
#define Morse_END   0x1A    // 0001 1010 = DI-DAH-DI-DAH (same as Morse_BEG)
#define Morse_PB	Ex80    // Pushbuton pressed code
#define Morse_NULL  0x01    // Empty return from GETCW
#define Morse_SP    0x1C    // Word space
#define Morse_PAD   0x2C    // 1/2 Padding

#define Morse_EXC   0x00    // ! 0001 0111 = DAH-DAH-DAH-DIT (MN)
#define Morse_2QU   0x52    // " 0101 0010 = DI-DAH-DI-DI-DAH-DIT (RR)
#define Morse_DOL   0xC8    // $ 1100 1000 = DI-DI-DI-DAH-DI-DI-DAH (SX)
#define Morse_1QU   0x5E    // ' 0101 1110 = DI-DAH-DAH-DAH-DAH-DIT (WG)
#define Morse_LPR   0x2D    // ( 0010 1101 = DAH-DI-DAH-DAH-DIT (KN)
#define Morse_RPR   0x6D    // ) 0110 1101 = DAH-DI-DAH-DAH-DI-DAH (KN)
#define Morse_PLS   0x2A    // + 0010 1010 = DI-DAH-DI-DAH-DIT (AR)
#define Morse_AT    0x56    // @ 0101 0110 = DI-DAH-DAH-DI-DAH-DIT (AC)
#define Morse_DSH   0x61    // - 0110 0001 = DAH-DI-DI-DI-DI-DAH (DU)
#define Morse_EQU   0x31    // = 0011 0001 = DAH-DI-DI-DI-DAH (BT)
#define Morse_COL   0x47    // : 0100 0111 = DAH-DAH-DAH-DI-DI-DIT (OS)
#define Morse_SCL   0x55    // ; 0101 0101 = DAH-DI-DAH-DI-DAH-DIT (KR)
#define Morse_UN    0x6C    // ; 0110 1100 = DI-DI-DAH-DAH-DI-DAH (UK)

#define Morse_0     0x3F    // 0 0011 1111 = DAH-DAH-DAH-DAH-DAH
#define Morse_1     0x3E    // 1 0011 1110 = DI-DAH-DAH-DAH-DAH
#define Morse_2     0x3C    // 2 0011 1100 = DI-DI-DAH-DAH-DAH
#define Morse_3     0x38    // 3 0011 1000 = DI-DI-DI-DAH-DAH
#define Morse_4     0x30    // 4 0011 0000 = DI-DI-DI-DI-DAH
#define Morse_5     0x20    // 5 0010 0000 = DI-DI-DI-DI-DIT
#define Morse_6     0x21    // 6 0010 0001 = DAH-DI-DI-DI-DIT
#define Morse_7     0x23    // 7 0010 0011 = DAH-DAH-DI-DI-DIT
#define Morse_8     0x27    // 8 0010 0111 = DAH-DAH-DAH-DI-DIT
#define Morse_9     0x2F    // 9 0010 1111 = DAH-DAH-DAH-DAH-DIT
#define Morse_AR    0x2A    //   0010 1010 = DI-DAH-DI-DAH-DIT
#define Morse_SK    0x68    //   0110 1000 = DI-DI-DI-DAH-DI-DAH
#define Morse_PER   0x6A    //   0110 1010 = DI-DAH-DI-DAH-DI-DAH
#define Morse_COM   0x73    //   0111 0011 = DAH-DAH-DI-DI-DAH-DAH
#define Morse_KN	0x2D    //   0010 1101 = DAH-DI-DAH-DAH-DIT
#define Morse_BT    0x31    //   0011 0001 = DAH-DI-DI-DI-DAH
#define Morse_AS    0x22    //   0010 0010 = DI-DAH-DI-DI-DI
#define Morse_QUE   0x4C    // ? 0100 1100 = DI-DI-DAH-DAH-DI-DIT
#define Morse_DN    0x29    //   0010 1001 = DAH-DI-DI-DAH-DIT
#define Morse_A     0x06    // A 0000 0110 = DI-DAH
#define Morse_B     0x11    // B 0001 0001 = DAH-DI-DI-DIT
#define Morse_C     0x15    // C 0001 0101 = DAH-DI-DAH-DIT
#define Morse_D     0x09    // D 0000 1001 = DAH-DI-DIT
#define Morse_E     0x02    // E 0000 0010 = DIT
#define Morse_F     0x14    // F 0001 0100 = DI-DI-DAH-DIT
#define Morse_G     0x0B    // G 0000 1011 = DAH-DAH-DIT
#define Morse_H     0x10    // H 0001 0000 = DI-DI-DI-DIT
#define Morse_I     0x04    // I 0000 0100 = DI-DIT
#define Morse_J     0x1E    // J 0001 1110 = DI-DAH-DAH-DAH
#define Morse_K     0x0D    // K 0000 1101 = DAH-DI-DAH
#define Morse_L     0x12    // L 0001 0010 = DI-DAH-DI-DIT
#define Morse_M     0x07    // M 0000 0111 = DAH-DAH
#define Morse_N     0x05    // N 0000 0101 = DAH-DIT
#define Morse_O     0x0F    // O 0000 1111 = DAH-DAH-DAH
#define Morse_P     0x16    // P 0001 0110 = DI-DAH-DAH-DIT
#define Morse_Q     0x1B    // Q 0001 1011 = DAH-DAH-DI-DAH
#define Morse_R     0x0A    // R 0000 1010 = DI-DAH-DIT
#define Morse_S     0x08    // S 0000 1000 = DI-DI-DIT
#define Morse_T     0x03    // T 0000 0011 = DAH
#define Morse_U     0x0C    // U 0000 1100 = DI-DI-DAH
#define Morse_V     0x18    // V 0001 1000 = DI-DI-DI-DAH
#define Morse_W     0x0E    // W 0000 1110 = DI-DAH-DAH
#define Morse_X     0x19    // X 0001 1001 = DAH-DI-DI-DAH
#define Morse_Y     0x1D    // Y 0001 1101 = DAH-DI-DAH-DAH
#define Morse_Z     0x13    // Z 0001 0011 = DAH-DAH-DI-DIT

#define Morse_AK    0x36    //   0011 0110 = ARING: DI-DAH-DAH-DI-DAH
#define Morse_MN    0x17    //   0001 0111 = OSLSH: DAH-DAH-DAH-DIT
#define Morse_AA    0x1A    //   0001 1010 = AE:    DI-DAH-DI-DAH
#define Morse_AC    0x56    //   0101 0110 = AC:    DI-DAH-DAH-DI-DAH-DIT

#define Morse_SN    0x28    //
#define Morse_DM    0x39    //
#define Morse_MM    0x1F    //
#define Morse_AC    0x56    //
#define Morse_AK    0x36    //
#define Morse_OA    0x37    //
#define Morse_SS    0x40    //
#define Morse_KA    0x35    //
#define Morse_RN    0x2A    //
#define Morse_RA    0x32    //
#define Morse_MU    0x33    //
#define Morse_MW    0x3B    //
#define Morse_ZA    0x63    //
#define Morse_ZW    0xE3    //
#define Morse_VG    0xB8    //
#define Morse_SZ    0x98    //
#define Morse_GO    0x7B    //
#define Morse_WG    0x5E    //
#define Morse_OG    0x5F    //
#define Morse_UG    0x5C    //
#define Morse_IM    0x1C    //
#define Morse_BA    0x61    //

// MARK: - WinKeyer ASCII to Morse conversion table -
unsigned char const ASCII_TO_MORSE[96] =
{
 Morse_SP,     0x01,Morse_2QU,     0x01,Morse_DOL,     0x01,     0x01,Morse_1QU,  /*  !"#$%&' */
Morse_LPR,Morse_RPR,     0x01,Morse_PLS,Morse_COM,Morse_DSH,Morse_PER, Morse_DN,  /* ()*+,-./ */
  Morse_0,  Morse_1,  Morse_2,  Morse_3,  Morse_4,  Morse_5,  Morse_6,  Morse_7,  /* 01234567 */
  Morse_8,  Morse_9, Morse_KN, Morse_AA, Morse_AR,Morse_EQU, Morse_SK,Morse_QUE,  /* 89:;<=>? */
 Morse_AC,  Morse_A,  Morse_B,  Morse_C,  Morse_D,  Morse_E,  Morse_F,  Morse_G,  /* @ABCDEFG */
  Morse_H,  Morse_I,  Morse_J,  Morse_K,  Morse_L,  Morse_M,  Morse_N,  Morse_O,  /* HIJKLMNO */
  Morse_P,  Morse_Q,  Morse_R,  Morse_S,  Morse_T,  Morse_U,  Morse_V,  Morse_W,  /* PQRSTUVW */
  Morse_X,  Morse_Y,  Morse_Z, Morse_AS, Morse_DN, Morse_KN,     0x01,     0x01,  /* XYZ[\]^_ */
     0x01,  Morse_A,  Morse_B,  Morse_C,  Morse_D,  Morse_E,  Morse_F,  Morse_G,  /* `abcdefg */
  Morse_H,  Morse_I,  Morse_J,  Morse_K,  Morse_L,  Morse_M,  Morse_N,  Morse_O,  /* hijklmno */
  Morse_P,  Morse_Q,  Morse_R,  Morse_S,  Morse_T,  Morse_U,  Morse_V,  Morse_W,  /* pqrstuvw */
  Morse_X,  Morse_Y,  Morse_Z,     0x01,Morse_PAD,     0x01,     0x01,     0x01   /* wyyz{|}  */
};

// MARK: - WinKeyer 2 EEPROM struct -
typedef struct {
    unsigned char magic;     // 0x00
    unsigned char modereg;   // 0x01
    unsigned char oprrate;   // 0x02
    unsigned char stconst;   // 0x03
    unsigned char weight;    // 0x04
    unsigned char lead_time; // 0x05
    unsigned char tail_time; // 0x06
    unsigned char minwpm;    // 0x07
    unsigned char wpmrange;  // 0x08
    unsigned char xtnd;      // 0x09
    unsigned char kcomp;     // 0x0a
    unsigned char farnswpm;  // 0x0b
    unsigned char sampadj;   // 0x0c
    unsigned char ratio;     // 0x0d
    unsigned char pincfg;    // 0x0e
    unsigned char k12mode;   // 0x0f
    unsigned char cmdwpm;    // 0x10
    unsigned char freeptr;   // 0x11
    unsigned char msgptr1;   // 0x12
    unsigned char msgptr2;
    unsigned char msgptr3;
    unsigned char msgptr4;
    unsigned char msgptr5;
    unsigned char msgptr6;
    unsigned char msgs[232]; // 0x18
} WinkeyerEEPROM;

#pragma mark static data

@implementation WinKeyerUsb

// MARK: - Initialization -

static WinKeyerUsb* sharedInstance = nil;
+ (WinKeyerUsb*)sharedInstance
{
    static dispatch_once_t singletonPredicate;
    dispatch_once(&singletonPredicate, ^{
        if (!sharedInstance) sharedInstance = [WinKeyerUsb new];
    });
    return sharedInstance;
}

- (instancetype)init
{
    if (sharedInstance != nil) {
        return sharedInstance;
    } else if ((self = [super init])) {
        sharedInstance = self;
        // Characters recognized by WinKeyer
        // Letters
        NSMutableCharacterSet* aSet = [[NSCharacterSet uppercaseLetterCharacterSet] mutableCopy];
        // Numbers
        [aSet formUnionWithCharacterSet:[NSCharacterSet decimalDigitCharacterSet]];
        // Punctuation
        [aSet addCharactersInString:@",.?;:"];
        // Symbols
        [aSet addCharactersInString:@"!#$@^*()-=+[]\\;:'\"<>/"];
        // Space
        [aSet addCharactersInString:@" "];

        // SkookumMorse special characters
        [aSet addCharactersInString:@"|"]; // '|' does half-space in messages and isn't echoed
        [aSet addCharactersInString:@"`~%{"]; // my call, their call, their name, merge
        self.morseKeyboardInvertedCharacterSet = [[aSet copy] invertedSet];
    }
    return self;
}

// MARK: - Actions -
- (IBAction)stopMessageSending:(id)sender
{
    if ([self isConnected] && [winKeyerSerialPort isOpen]) {
        [self putToWinKeyer:byteToData(kWKImmediateClearBufferCommand)];
        //NSLog(@"Told WinKeyer to stop sending");
        self.tuneState = NO;
    }
    [self clearSending];
}

- (IBAction)sendMessage:(id)sender
{
    NSInteger tag = [sender tag];
    NSString* key = nil;
    switch (tag) {
        case 0:
            key = kMessage1PreferenceKey;
            break;
        case 1:
            key = kMessage2PreferenceKey;
            break;
        case 2:
            key = kMessage3PreferenceKey;
            break;
        case 3:
            key = kMessage4PreferenceKey;
            break;
        case 4:
            key = kMessage5PreferenceKey;
            break;
        case 5:
            key = kMessage6PreferenceKey;
            break;
        case 6:
            key = kMessage7PreferenceKey;
            break;
        case 7:
            key = kMessage8PreferenceKey;
            break;
    }
    NSString* message = [self substituteTokens:[[NSUserDefaults standardUserDefaults] stringForKey:key]];
    [self appendMessage:message];
}

- (IBAction)pause:(NSButton*)sender
{
    unsigned char command[2] = {kWKImmediateSetPauseStateCommand, 0x00};
    if (sender.state == NSOnState) {
        command[1] = 0x01;
    }
    [self putToWinKeyer:bytesToData(command, 2)];
    if (sender.state == NSOffState) {
        [self send];
    }
}

- (IBAction)tune:(id)sender
{
    unsigned char command[2] = {kWKImmediateKeyImmediateCommand, 0x00};
    if ([sender state] == NSOnState) {
        command[1] = 0x01;
    }
    [self putToWinKeyer:bytesToData(command, 2)];
}

// MARK: - WinKeyer Setttings Actions -

- (IBAction)changeWeight:(id)sender
{
    const unsigned char weight = [[NSUserDefaults standardUserDefaults] integerForKey:kWKWeightingPreferenceKey] + 50;
    const unsigned char command[2] = {kWKImmediateSetWeightingCommand, weight};
    [self putToWinKeyer:bytesToData(command, 2)];
}

- (IBAction)changeRatio:(id)sender
{
    const unsigned char ratio = [[NSUserDefaults standardUserDefaults] doubleForKey:kWKRatioPreferenceKey] / 0.06;
    const unsigned char command[2] = {kWKImmediateSetDitDahRatioCommand, ratio};
    [self putToWinKeyer:bytesToData(command, 2)];
}

- (IBAction)changeFirstExtension:(id)sender
{
    const unsigned char extension = [[NSUserDefaults standardUserDefaults] doubleForKey:kWKFirstExtensionPreferenceKey];
    const unsigned char command[2] = {kWKImmediateSetFirstExtensionCommand, extension};
    [self putToWinKeyer:bytesToData(command, 2)];
}

- (IBAction)changeKeyingCompensation:(id)sender
{
    const unsigned char compensation = [[NSUserDefaults standardUserDefaults] doubleForKey:kWKKeyingCompensationPreferenceKey];
    const unsigned char command[2] = {kWKImmediateSetKeyingCompensationCommand, compensation};
    [self putToWinKeyer:bytesToData(command, 2)];
}

- (IBAction)changeWinKeyer2Mode:(id)sender
{
    unsigned char modeRegister = 0x00;
    BOOL disablePaddleWatchDog = [[NSUserDefaults standardUserDefaults] boolForKey:kWKDisablePaddleWatchdogPreferenceKey];
    if (disablePaddleWatchDog) modeRegister |= 0x80;

    BOOL paddleEchoback = [[NSUserDefaults standardUserDefaults] boolForKey:kWKPaddleEchobackPreferenceKey];
    if (paddleEchoback) modeRegister |= 0x40;

    PaddleModeType paddleMode = [[NSUserDefaults standardUserDefaults] integerForKey:kWKKeyModePreferenceKey];
    if (paddleMode == PaddleModeIambicA) modeRegister |= 0x10;
    else if (paddleMode == PaddleModeUltimatic) modeRegister |= 0x20;
    else if (paddleMode == PaddleModeBug) modeRegister |= 0x30;

    BOOL paddleSwap = [[NSUserDefaults standardUserDefaults] boolForKey:kWKPaddleSwapPreferenceKey];
    if (paddleSwap) modeRegister |= 0x08;

//    BOOL serialEchoback = [[NSUserDefaults standardUserDefaults] boolForKey:kWKSerialEchobackPreferenceKey];
    if (YES) modeRegister |= 0x04;

    BOOL autospace = [[NSUserDefaults standardUserDefaults] boolForKey:kWKAutospacePreferenceKey];
    if (autospace) modeRegister |= 0x02;

    BOOL contestSpacing = [[NSUserDefaults standardUserDefaults] boolForKey:kWKContestSpacingPreferenceKey];
    if (contestSpacing) modeRegister |= 0x01;

    const unsigned char command[2] = {kWKImmediateSetWinkeyer2ModeCommand, modeRegister};
    [self putToWinKeyer:bytesToData(command, 2)];
}

- (IBAction)changeSidetone:(id)sender
{
    uint8 sidetoneRegister = 0x00;
    if (self.isVersion2) {
        sidetoneRegister = [self encodeWK2SidetoneRegister];
    } else {
        sidetoneRegister = [self encodeWK3SidetoneRegister];
    }
    const uint8 command[2] = {kWKImmediateSidetoneControlCommand, sidetoneRegister};
    [self putToWinKeyer:bytesToData(command, 2)];
    if (!self.isVersion2) {
        uint8 extensionRegister = [self encodeExtension2Register];
        const uint8 adminCommand[3] = {kWKAdminWriteX2MODECommand[0], kWKAdminWriteX2MODECommand[1], extensionRegister};
        [self putToWinKeyer:bytesToData(adminCommand, 3)];
    }
}

- (IBAction)changePTT:(id)sender
{
    const unsigned char pttLeadTime = [[NSUserDefaults standardUserDefaults] integerForKey:kWKPTTLeadTimePreferenceKey];
    const unsigned char pttTailTime = [[NSUserDefaults standardUserDefaults] integerForKey:kWKPTTTailTimePreferenceKey];
    const unsigned char command[3] = {kWKImmediateSetPTTLeadTailCommand, pttLeadTime, pttTailTime};
    [self putToWinKeyer:bytesToData(command, 3)];
}

- (IBAction)changeSpeedPot:(id)sender
{
    const unsigned char potMinimum = [[NSUserDefaults standardUserDefaults] integerForKey:kWKSpeedPotMinimumPreferenceKey];
    const unsigned char potRange = [[NSUserDefaults standardUserDefaults] integerForKey:kWKSpeedPotRangePreferenceKey];
    const unsigned char command[4] = {kWKImmediateSetupSpeedPotCommand, potMinimum, potRange, 0xFF};
    [self putToWinKeyer:bytesToData(command, 4)];
    const unsigned char potLockCommand[2] = {kWKImmediateSetWPMSpeedCommand, 0};
    [self putToWinKeyer:bytesToData(potLockCommand, 2)];
    // Cause update of speed
    [self putToWinKeyer:byteToData(kWKImmediateGetSpeedPotCommand)];
}

- (IBAction)changeFarnsworthSpeed:(id)sender
{
    const unsigned char farnsworthSpeed = [[NSUserDefaults standardUserDefaults] integerForKey:kWKFarnsworthSpeedPreferenceKey];
    const unsigned char command [2] = {kWKImmediateSetFarnsworthWPMCommand, farnsworthSpeed};
    [self putToWinKeyer:bytesToData(command, 2)];
}

- (IBAction)changePaddleSwitchpoint:(id)sender
{
    const unsigned char paddleSwitchpoint = [[NSUserDefaults standardUserDefaults] integerForKey:kWKPaddleSwitchpointPreferenceKey];
    const unsigned char command[2] = {kWKImmediateSetPaddleSwitchpointCommand, paddleSwitchpoint};
    [self putToWinKeyer:bytesToData(command, 2)];
}

- (IBAction)changeEnableDitDahMemory:(id)sender
{
    BOOL enableDitDahMemory = [[NSUserDefaults standardUserDefaults] boolForKey:kWKEnableDitDahMemoryPreferenceKey];
    if (enableDitDahMemory) {
        [self changePaddleSwitchpoint:sender];
    } else {
        const unsigned char command[2] = {kWKImmediateSetPaddleSwitchpointCommand, 0x00};
        [self putToWinKeyer:bytesToData(command, 2)];
    }
}

- (IBAction)changePinConfiguration:(id)sender
{
    unsigned char pinConfiguration = 0x00;
    UltimaticPriorityType ditdahPriority = [[NSUserDefaults standardUserDefaults] integerForKey:kWKDitDahPriorityPreferenceKey];
    if (ditdahPriority == UltimaticPriorityDah) pinConfiguration |= 0x40;
    else if (ditdahPriority == UltimaticPriorityDit) pinConfiguration |= 0x80;

    HangTimeType paddleHangTime = [[NSUserDefaults standardUserDefaults] integerForKey:kWKPaddleHangPreferenceKey];
    if (paddleHangTime == HangTimeShort) pinConfiguration |= 0x10;
    else if (paddleHangTime == HangTimeLong) pinConfiguration |= 0x20;
    else if (paddleHangTime == HangTimeLongest) pinConfiguration |= 0x30;

    BOOL keyOut1Enable = [[NSUserDefaults standardUserDefaults] boolForKey:kWKKeyOut1EnablePreferenceKey];
    if (keyOut1Enable) pinConfiguration |= 0x04;

    BOOL keyOut2Enable = [[NSUserDefaults standardUserDefaults] boolForKey:kWKKeyOut2EnablePreferenceKey];
    if (keyOut2Enable) pinConfiguration |= 0x08;

    BOOL sidetoneEnable = [[NSUserDefaults standardUserDefaults] boolForKey:kWKSidetoneEnablePreferenceKey];
    if (sidetoneEnable) pinConfiguration |= 0x02;

    BOOL PTTEnable = [[NSUserDefaults standardUserDefaults] boolForKey:kWKPTTEnablePreferenceKey];
    if (PTTEnable) pinConfiguration |= 0x01;

    const unsigned char command[2] = {kWKImmediateSetPinConfigurationCommand, pinConfiguration};
    [self putToWinKeyer:bytesToData(command, 2)];
}

// MARK: - Public Methods -
- (BOOL)connect:(NSError**)error
{
    NSString* portName;
    if (![self determinePortName:&portName error:error]) return NO;
    
    if (![self determineSerialPortForName:portName error:error]) return NO;
    
    //NSLog(@"serial port properties: %@", [winKeyerSerialPort properties]);
    
//TODO    if (![self isSerialPortAvailable:error]) return NO;
    
    if (![self configureSerialPort:error]) return NO;
    
    if (![self openSerialPort:error]) return NO;

    if (![self syncWinKeyer:error]) return NO;
    
    winKeyerSerialPort.delegate = self;
    
    [self openHostMode];
    
    return YES;
}

- (BOOL)isConnected
{
    return winKeyerSerialPort != nil;
}

- (BOOL)disconnect:(NSError**)error
{
    BOOL success = YES;
    // Clobber whatever is going on
    [self clearSending];
    self.versionString = nil;
    self.speedString = nil;
    
    if ([self isConnected]) {
        if ([winKeyerSerialPort isOpen]) {
            sleep(1);
            success = [self closeHostMode:error];
            sleep(1); // Don't clobber port too quickly after above commands
            [winKeyerSerialPort close];
        }
        winKeyerSerialPort = nil;
    }
    return success;
}

- (void)putToWinKeyer:(NSData*)data
{
    if ([self isConnected]) {
        // NSLog(@"Put to WinKeyer data = %@", data);
        BOOL sendOK = [winKeyerSerialPort sendData:data];
        if (!sendOK) {
            NSLog(@"putToWinKeyer failed for data = %@", data);
        }
    }
}

// MARK: - ORSSerial delegate methods -

- (void)serialPort:(ORSSerialPort*)serialPort didReceiveResponse:(NSData*)responseData toRequest:(ORSSerialRequest*)request
{
    NSString* commandName = request.userInfo[@"CommandName"];
    //    NSLog(@"%s %@", __PRETTY_FUNCTION__, commandName);
    if ([commandName isEqualToString:@"HostOpen"]) {
        uint8 versionNumber = ((uint8*)responseData.bytes)[0];
        NSInteger majorVersion = versionNumber / 10;
        NSInteger minorVersion = versionNumber - 10 * majorVersion;
        if (majorVersion >= 3 && majorVersion < 4) {
            self.isVersion2 = NO;
            self.versionString = [NSString stringWithFormat:@"WinKeyer 3 version %1ld.%1ld", majorVersion, minorVersion];
            [self readMinorVersion];
        } else if (majorVersion >= 2 && majorVersion < 3) {
            self.isVersion2 = YES;
            self.versionString = [NSString stringWithFormat:@"WinKeyer 2 version %1ld.%1ld", majorVersion, minorVersion];
            [self echoTest];
        } else {
            NSString* errorDescription = [NSString stringWithFormat:@"WinKeyer returned unsupported version: %ld.", majorVersion];
            NSString* errorDetail = @"Confirm that the WinKeyer version is 20 or newer.";
            NSDictionary* errorDict = @{NSLocalizedDescriptionKey: errorDescription, NSLocalizedRecoverySuggestionErrorKey: errorDetail};
            NSError* error = [NSError errorWithDomain:@"com.k1gq.macwinkeyer" code:1 userInfo:errorDict];
            [[NSUserDefaults standardUserDefaults] setObject:kWKNoPortString forKey:kWKPortNamePreferenceKey];
            [NSApp presentError:error];
            [winKeyerSerialPort close];
            self.versionString = @"";
        }
    } else if ([commandName isEqualToString:@"ReadMinorVersion"]) {
        // TODO: Debug inconsistent minorVersion
        uint8 versionNumber = ((uint8*)responseData.bytes)[0];
        NSInteger majorVersion = versionNumber / 10;
        NSInteger minorVersion = versionNumber - 10 * majorVersion;
        self.versionString = [self.versionString stringByAppendingFormat:@".%1ld", minorVersion];
        [self readType];
    } else if ([commandName isEqualToString:@"ReadType"]) {
        WinKeyerType type = ((uint8*)responseData.bytes)[0];
        if (type == WinKeyerTypeMINI) {
            self.versionString = [self.versionString stringByReplacingOccurrencesOfString:@"WinKeyer 3" withString:@"WKmini"];
        }
        [self echoTest];
    } else if ([commandName isEqualToString:@"EchoTest"]) {
        uint8 echo = ((uint8*)responseData.bytes)[0];
        if (echo == 0x52) {
            // Contrary to "Winkeyer2 IC v22 Interface and Operation Manual 6/6/2008", the
            // Admin Set WK2 Mode command is not recognized unless set *after* Host Mode is open.
            if (self.isVersion2) {
                [self putToWinKeyer:bytesToData(kWKAdminSetWK2ModeCommand, 2)];
            } else {
                [self putToWinKeyer:bytesToData(kWKAdminSetWK2ModeCommand, 2)];
                [self putToWinKeyer:bytesToData(kWKAdminSetWK3ModeCommand, 2)];
            }
            [self loadDefaults];
            [self putToWinKeyer:byteToData(kWKImmediateGetSpeedPotCommand)];
            _isHostMode = YES;
            //        if (self.version31) {
            //            [self changeFirstExtension:nil];
            //        }
            //        self.hostMode = YES;
            //        [self changeStatusString];
        } else {
            NSString* errorDescription = [NSString stringWithFormat:@"Echo test failed."];
            NSString* errorDetail = [NSString stringWithFormat:@"Expected 'R', got %c", echo];
            NSDictionary* errorDict = @{NSLocalizedDescriptionKey: errorDescription, NSLocalizedRecoverySuggestionErrorKey: errorDetail};
            NSError* error = [NSError errorWithDomain:@"com.k1gq.macwinkeyer" code:1 userInfo:errorDict];
            [[NSUserDefaults standardUserDefaults] setObject:kWKNoPortString forKey:kWKPortNamePreferenceKey];
            [NSApp presentError:error];
            [winKeyerSerialPort close];
            self.versionString = @"";
        }
    } else {
        NSLog(@"%s unhandled request response name: %@", __PRETTY_FUNCTION__, commandName);
    }
}

// Called when data arrives from WinKeyer USB
- (void)serialPort:(ORSSerialPort*)serialPort didReceiveData:(NSData*)data
{
//    NSLog(@"%s: port name = %@, data = %@", __PRETTY_FUNCTION__, serialPort.name, data);
    if (serialPort != winKeyerSerialPort) {
        NSLog(@"Winkeyer didReceiveData called by a stranger!");
        NSLog(@"sourcePort address: %p, name: %@", serialPort, serialPort.name);
    }
    
    _wkStatus = 0x00;
    if ([data length] == 0) {
        NSLog(@"WinKeyer port was somehow closed!");
        return;
    }
    // NSLog(@"serialPortReadData got %d bytes", [input length]);
    for (NSUInteger i = 0; i < [data length]; ++i) {
        unsigned char aByte = ((unsigned char*)[data bytes])[0];
        //NSLog(@"  WinKeyer said: %x (%c)", aByte, aByte);
        if (aByte == 0xFF) {
            NSLog(@"  error byte");
            
        } else if ((aByte & 0xE0) == 0xC0) {
            _wkStatus = aByte;
            if (aByte & 0x08) { // push button status byte
                // NSLog(@"  button pressed status byte");
                [self handleButtonPressed];
            } else { // WK2 status byte
                //NSLog(@"  WK2 status byte");
                if (!(aByte & 0x1F)) {
                    //NSLog(@"    Nothing to report");
//                    if (self.pauseState) {
//                        self.pauseState = NSOffState;
//                    }
                    if (self.beaconState) {
                        double beaconInterval = [[NSUserDefaults standardUserDefaults] doubleForKey:kBeaconIntervalPreferenceKey];
                        _beaconTimer = [NSTimer scheduledTimerWithTimeInterval:beaconInterval
                                                                        target:self
                                                                      selector:@selector(repeatMessage:)
                                                                      userInfo:nil
                                                                       repeats:NO];
                    }
                    [self send];
                }
//                BOOL bufferNearlyFull = (aByte & 0x01);
//                if (bufferNearlyFull) {
//                    //NSLog(@"    WinKeyer said buffer 2/3 full");
//                    self.pauseState = NSOnState;
//                }
                BOOL paddleBreakin = (aByte & 0x02);
                if (paddleBreakin) {
                    //NSLog(@"    WinKeyer said paddle breakin");
                    [self clearSending];
                }
                BOOL busy = (aByte & 0x04);
                if (busy) {
                    //NSLog(@"    WinKeyer said busy");
                    if (self.beaconState) {
                        [_beaconTimer invalidate];
                    }
                }
                BOOL waiting = (aByte & 0x10);
                if (waiting) {
                    //NSLog(@"    WinKeyer said waiting");
                }
            }
            
        } else if ((aByte & 0xE0) == 0x80) {
            //NSLog(@"  speed pot changed");
            NSUserDefaults* defaults = [NSUserDefaults standardUserDefaults];
            NSInteger speedPotMinimum = [defaults integerForKey:kWKSpeedPotMinimumPreferenceKey];
            // Get Speed Pot response
            NSInteger response = aByte & 0x3F;
            NSInteger potSpeedWPM = response + speedPotMinimum;
            self.speedString = [NSString stringWithFormat:@"%ld WPM", potSpeedWPM];
            
            
        } else {
            //NSLog(@"  serial echo: %c", aByte);
            // Assume it's the expected character
            [self send];
        }
    }
}

- (void)serialPortWasRemovedFromSystem:(ORSSerialPort *)serialPort
{
//    NSLog(@"%s", __PRETTY_FUNCTION__);
    winKeyerSerialPort = nil;
}

// MARK: - NSTextDelegate methods -
// For keyboard input text view

- (BOOL)textShouldBeginEditing:(NSText*)aTextObject
{
    return YES;
}

- (BOOL)textShouldEndEditing:(NSText*)aTextObject
{
    return YES;
}

- (void)textDidBeginEditing:(NSNotification*)aNotification
{
}

- (void)textDidEndEditing:(NSNotification*)aNotification
{
}

- (void)textDidChange:(NSNotification*)aNotification
{
    //NSLog(@"WinKeyerUsb:textDidChange");
    if ([self.keyboardBuffer length] == 0) {
        _outputBufferIndex = 0;
        return;
    }
    
    unichar c = 0x0000;
    NSMutableString* candidateString = [NSMutableString stringWithString:[self.keyboardBuffer uppercaseString]];
    for (NSUInteger i = 0; i < [candidateString length]; ++i) {
        c = [candidateString characterAtIndex:i];
        if ([self.morseKeyboardInvertedCharacterSet characterIsMember:c]) {
            [candidateString deleteCharactersInRange:NSMakeRange(i, 1)];
        }
    }
    self.keyboardBuffer = candidateString;
    self.keyboardBuffer = [self substituteTokens:self.keyboardBuffer];
    if (_outputBufferIndex >= [self.keyboardBuffer length]) {
        _outputBufferIndex = [self.keyboardBuffer length];
    }
    if (!self.pauseState && !_sendingState) [self send];
}

// MARK: - ORSSerial Admin Commands -

- (void)openHostMode
{
    // This sets off a series of requests and commands
    //  HostOpen, returns version
    //  If version 3
    //    ReadMinorVersion
    //    ReadType
    //    EchoTest
    //  Else
    //    EchoTest
    //  If version 3
    //    AdminSetWK2Mode
    //    AdminSetWK3Mode
    //  Else
    //    AdminSetWK2Mode
    //  loadDefaults
    //  GetSpeedPot
    //  set _isHostMode
    if (!_isHostMode) [self hostOpen];
}

- (BOOL)closeHostMode:(NSError**)error
{
    if (!_isHostMode) return YES; // already closed
    [self putToWinKeyer:bytesToData(kWKAdminHostCloseCommand, 2)];
    //NSLog(@"Did close Host Mode");
    _isHostMode = NO;
    return YES;
}

// MARK: - Send -
- (void)send
{
    if (self.pauseState) return;
    _currentBufferIndex = _outputBufferIndex;
    if (_outputBufferIndex < [self.keyboardBuffer length]) {
        const char c = [[self.keyboardBuffer uppercaseString] UTF8String][_outputBufferIndex];
        _outputBufferIndex += 1;
        _sendingState = YES;
        switch (c) {
            case '{':
                NSLog(@"Merge letters");
                if (_outputBufferIndex + 1 < [self.keyboardBuffer length]) {
                    const char mergeLeft = [[self.keyboardBuffer uppercaseString] UTF8String][_outputBufferIndex + 0];
                    const char mergeRight = [[self.keyboardBuffer uppercaseString] UTF8String][_outputBufferIndex + 1];
                    unsigned char wkCommand[3] = {kWKBufferedMergeLettersCommand, mergeLeft, mergeRight};
                    [self putToWinKeyer:bytesToData(wkCommand, 3)];
                    _outputBufferIndex += 2;
                    // Not echoed
                    [self send];
                }
                break;
            case '|':
                // Not echoed by WinKeyer, so won't get called back by serialReadData
                [self putToWinKeyer:byteToData(c)];
                [self send];
                break;
            default:
                // serialReadData calls send after character is echoed.
                [self putToWinKeyer:byteToData(c)];
        }
    } else {
        _sendingState = NO;
    }
    [self updateKeyboardTextViewBackgroundColor];
}

- (void)updateKeyboardTextViewBackgroundColor
{
    // Indicate sent (green) versus pending (yellow) characters with text background color
    if (!self.keyboardBuffer) return;
    NSMutableAttributedString *newAttrString = [[NSMutableAttributedString alloc] initWithString:self.keyboardBuffer.uppercaseString];
    NSFont *font = [NSFont fontWithName:@"Monaco" size:14.0];
    [newAttrString addAttribute:NSFontAttributeName value:font range:NSMakeRange(0, newAttrString.length)];
    NSDictionary* attrsDefaultTextColor = @{NSForegroundColorAttributeName : NSColor.textColor};
    [newAttrString addAttributes:attrsDefaultTextColor range:NSMakeRange(0, newAttrString.length)];
    NSDictionary *attrsGreenText = @{ NSBackgroundColorAttributeName : NSColor.systemGreenColor };
    if (_sendingState) {
        [newAttrString addAttributes:attrsGreenText range:NSMakeRange(0, _currentBufferIndex)];
        NSDictionary *attrsYellowText = @{ NSBackgroundColorAttributeName : NSColor.systemYellowColor };
        [newAttrString addAttributes:attrsYellowText range:NSMakeRange(_currentBufferIndex, _outputBufferIndex - _currentBufferIndex)];
    } else {
        [newAttrString addAttributes:attrsGreenText range:NSMakeRange(0, newAttrString.length)];
    }
    SkookumMorseDelegate* appDelegate = (SkookumMorseDelegate*)NSApp.delegate;
    NSTextView* morseKeyboardTextView = appDelegate.morseKeyboardTextView;
    [[morseKeyboardTextView textStorage] setAttributedString:newAttrString];
}

- (void)clearSending
{
    _outputBufferIndex = 0;
    _sendingState = NO;
    self.keyboardBuffer = @"";
    self.beaconState = NSOffState;
}

- (void)appendMessage:(NSString*)aMessage
{
    if (nil == self.keyboardBuffer) {
        self.keyboardBuffer = [aMessage uppercaseString];
    } else {
        if ([self.keyboardBuffer length] > 0) {
            self.keyboardBuffer = [self.keyboardBuffer stringByAppendingString:@" "];
        }
        self.keyboardBuffer = [self.keyboardBuffer stringByAppendingString:[aMessage uppercaseString]];
    }
    [self send];
    if (self.beaconState) {
        _beaconMessage = aMessage;
    }
}

- (void)repeatMessage:(NSTimer*)aTimer
{
    if (self.beaconState) {
        [self appendMessage:_beaconMessage];
    }
}

- (NSString*)myCall
{
    return [[NSUserDefaults standardUserDefaults] stringForKey:kMyCallPreferenceKey];
}

- (NSString*)substituteTokens:(NSString*)message
{
    NSMutableString* result = [NSMutableString stringWithString:message];
    [result replaceOccurrencesOfString:@"`" 
                            withString:[self myCall] 
                               options:0 
                                 range:NSMakeRange(0, result.length)];
    NSString* hisCall = (self.hisCallsign ? self.hisCallsign : @"");
    [result replaceOccurrencesOfString:@"~"
                            withString:hisCall
                               options:0
                                 range:NSMakeRange(0, result.length)];
    NSString* name = (self.hisName ? self.hisName : @"");
    [result replaceOccurrencesOfString:@"%"
                            withString:name
                               options:0
                                 range:NSMakeRange(0, result.length)];
    return result;
}

// MARK: - WinKeyer Settings -

- (uint8)encodeModeRegister
{
    NSUserDefaults* prefs = [NSUserDefaults standardUserDefaults];
    uint8 modeRegister = 0x00;
//    BOOL paddleWatchDog = [preferences[PaddleWatchDog] boolValue];
//    if (paddleWatchDog) modeRegister |= 0x80;
//    BOOL paddleEchoback = [preferences[HostPaddleEchoback] boolValue];
//    if (paddleEchoback) modeRegister |= 0x40;
    PaddleModeType paddleMode = [prefs integerForKey:kWKKeyModePreferenceKey];
    switch (paddleMode) {
        case PaddleModeIambicB: break;
        case PaddleModeIambicA: modeRegister |= 0x10; break;
        case PaddleModeUltimatic: modeRegister |= 0x20; break;
        case PaddleModeBug: modeRegister |= 0x30; break;
    }
    BOOL paddleSwap = [prefs boolForKey:kWKPaddleSwapPreferenceKey];
    if (paddleSwap) modeRegister |= 0x08;
//    BOOL serialEchoback = [preferences[HostSerialEchoback] boolValue];
    if (YES) modeRegister |= 0x04;
    BOOL autoSpace = [prefs boolForKey:kWKAutospacePreferenceKey];
    if (autoSpace) modeRegister |= 0x02;
    BOOL contestSpace = [prefs boolForKey:kWKContestSpacingPreferenceKey];
    if (contestSpace) modeRegister |= 0x01;
    return modeRegister;
}

- (uint8)encodePinConfigurationRegister
{
    NSUserDefaults* prefs = [NSUserDefaults standardUserDefaults];
    uint8 pinConfigurationRegister = 0x00;
    UltimaticPriorityType ultimaticPriority = [prefs integerForKey:kWKDitDahPriorityPreferenceKey];
    switch (ultimaticPriority) {
        case UltimaticPriorityNormal: break;
        case UltimaticPriorityDah: pinConfigurationRegister |= 0x40; break;
        case UltimaticPriorityDit: pinConfigurationRegister |= 0x80; break;
    }
    HangTimeType paddleHangTime = [prefs integerForKey:kWKPaddleHangPreferenceKey];
    switch (paddleHangTime) {
        case HangTimeShortest: break;
        case HangTimeShort: pinConfigurationRegister |= 0x10; break;
        case HangTimeLong: pinConfigurationRegister |= 0x20; break;
        case HangTimeLongest: pinConfigurationRegister |= 0x30; break;
    }
    BOOL key2Enable = [prefs boolForKey:kWKKeyOut2EnablePreferenceKey];
    if (key2Enable) pinConfigurationRegister |= 0x08;
    BOOL key1Enable = [prefs boolForKey:kWKKeyOut1EnablePreferenceKey];
    if (key1Enable) pinConfigurationRegister |= 0x04;
    BOOL sidetoneEnable = [prefs boolForKey:kWKSidetoneEnablePreferenceKey];
    if (sidetoneEnable) pinConfigurationRegister |= 0x02;
    BOOL pttEnable = [prefs boolForKey:kWKPTTEnablePreferenceKey];
    if (pttEnable) pinConfigurationRegister |= 0x01;
    return pinConfigurationRegister;
}

- (uint8)encodeWK2SidetoneRegister
{
    uint8 sidetoneRegister = 0x00;
    SidetoneFrequencyType sidetoneFrequency = [[NSUserDefaults standardUserDefaults] integerForKey:kWKSidetoneFrequencyPreferenceKey];
    switch (sidetoneFrequency) {
        case Sidetone4000Hz: sidetoneRegister |= 0x01; break;
        case Sidetone2000Hz: sidetoneRegister |= 0x02; break;
        case Sidetone1333Hz: sidetoneRegister |= 0x03; break;
        case Sidetone1000Hz: sidetoneRegister |= 0x04; break;
        case Sidetone800Hz: sidetoneRegister |= 0x05; break;
        case Sidetone666Hz: sidetoneRegister |= 0x06; break;
        case Sidetone571Hz: sidetoneRegister |= 0x07; break;
        case Sidetone500Hz: sidetoneRegister |= 0x08; break;
        case Sidetone444Hz: sidetoneRegister |= 0x09; break;
        case Sidetone400Hz: sidetoneRegister |= 0x0A; break;
    }
    BOOL paddleOnly = [[NSUserDefaults standardUserDefaults] boolForKey:kWKPaddleOnlySidetonePreferenceKey];
    if (paddleOnly) sidetoneRegister |= 0x80;
    return sidetoneRegister;
}

- (uint8)encodeWK3SidetoneRegister
{
    uint8 sidetoneRegister = 0x00;
    SidetoneFrequencyType sidetoneFrequency = [[NSUserDefaults standardUserDefaults] integerForKey:kWKSidetoneFrequencyPreferenceKey];
    switch (sidetoneFrequency) {
        case Sidetone4000Hz: sidetoneRegister = 65200/4000; break;
        case Sidetone2000Hz: sidetoneRegister = 65200/2000; break;
        case Sidetone1333Hz: sidetoneRegister = 65200/1333; break;
        case Sidetone1000Hz: sidetoneRegister = 65200/1000; break;
        case Sidetone800Hz: sidetoneRegister = 65200/800; break;
        case Sidetone666Hz: sidetoneRegister = 65200/666; break;
        case Sidetone571Hz: sidetoneRegister = 65200/571; break;
        case Sidetone500Hz: sidetoneRegister = 65200/500; break;
        case Sidetone444Hz: sidetoneRegister = 65200/444; break;
        case Sidetone400Hz: sidetoneRegister = 65200/400; break;
    }
    return sidetoneRegister;
}

- (uint8)encodeExtension2Register
{
    uint8 extensionRegister = 0x00;
//    BOOL enablePaddleStatus = [preferences[PaddleStatusEnable] boolValue];
//    if (enablePaddleStatus) extensionRegister |= 0x80;
//    BOOL fastStandaloneCommandResponse = [preferences[StandaloneFastCommand] boolValue];
//    if (NO) extensionRegister |= 0x40;
//    BOOL standaloneNCut = [preferences[SerialNumberCut9] boolValue];
//    if (NO) extensionRegister |= 0x20;
//    BOOL standaloneTCut = [preferences[SerialNumberCut0] boolValue];
//    if (NO) extensionRegister |= 0x10;
    BOOL paddleOnlySidetone = [[NSUserDefaults standardUserDefaults] boolForKey:kWKPaddleOnlySidetonePreferenceKey];
    if (paddleOnlySidetone) extensionRegister |= 0x08;
//    BOOL so2rMode = [preferences[PushToTalkSO2R] boolValue];
//    if (NO) extensionRegister |= 0x04;
//    BOOL paddleMute = [preferences[PaddleMute] boolValue];
//    if (NO) extensionRegister |= 0x02;
    return extensionRegister;
}

- (uint8)encodeWK3Extension1Register
{
//    NSUserDefaults* prefs = [NSUserDefaults standardUserDefaults];
    uint8 extensionRegister = 0x00;
//    uint8 letterSpaceAdjustment = [preferences[PaddleLetterSpace] integerValue];
//    extensionRegister = (letterSpaceAdjustment & 0x1F);
//    NSUInteger userOne = [preferences[StandaloneUser] integerValue];
//    if (userOne == 1) extensionRegister |= 0x80;
//    BOOL messageBankOne = [preferences[MessageBank] boolValue];
//    if (messageBankOne) extensionRegister |= 0x40;
//    BOOL fiftyPercentTune = [preferences[KeyTuneDutyCycle] boolValue];
//    if (fiftyPercentTune) extensionRegister |= 0x20;
    return extensionRegister;
}


- (void)loadDefaults
{
    [self changeWeight:self];
    [self changeRatio:self];
    [self changeFirstExtension:self];
    [self changeKeyingCompensation:self];
    [self changeWinKeyer2Mode:self];
    [self changeSidetone:self];
    [self changePTT:self];
    [self changeSpeedPot:self];
    [self changeFarnsworthSpeed:self];
    [self changePaddleSwitchpoint:self];
    [self changeEnableDitDahMemory:self];
    [self changePinConfiguration:self];

//    NSUserDefaults* prefs = [NSUserDefaults standardUserDefaults];
//    uint8 defaults[16];
//    defaults[0] = kWKImmediateLoadDefaultsCommand;
//    defaults[1] = [self encodeModeRegister];
////    if (prefs[SpeedPotLock]) {
//    defaults[2] = 0;
////    } else {
////        defaults[2] = [prefs[SpeedCommanded] integerValue];
////    }
//    if (self.isVersion2) {
//        defaults[3] = [self encodeWK2SidetoneRegister];
//    } else {
//        defaults[3] = [self encodeWK3SidetoneRegister];
//    }
//    defaults[4] = [prefs integerForKey:kWKWeightingPreferenceKey];
//    defaults[5] = [prefs integerForKey:kWKPTTLeadTimePreferenceKey];
//    defaults[6] = [prefs integerForKey:kWKPTTTailTimePreferenceKey];
//    defaults[7] = [prefs integerForKey:kWKSpeedPotMinimumPreferenceKey];
//    defaults[8] = [prefs integerForKey:kWKSpeedPotRangePreferenceKey];
//    if (self.isVersion2) {
//        defaults[9] = [prefs integerForKey:kWKFirstExtensionPreferenceKey];
//    } else {
//        defaults[9] = [self encodeExtension2Register];
//    }
//    defaults[10] = [prefs integerForKey:kWKKeyingCompensationPreferenceKey];
//    defaults[11] = [prefs integerForKey:kWKFarnsworthSpeedPreferenceKey];
//    defaults[12] = [prefs integerForKey:kWKPaddleSwitchpointPreferenceKey];
//    defaults[13] = [prefs integerForKey:kWKRatioPreferenceKey];
//    defaults[14] = [self encodePinConfigurationRegister];
//    if (self.isVersion2) {
//        defaults[15] = 0x00;
//    } else {
//        defaults[15] = [self encodeWK3Extension1Register];
//    }
//    [self putToWinKeyer:bytesToData(defaults, 16)];
}

- (void)handleButtonPressed
{
    NSString* message = nil;
    if (_wkStatus & 0x01) message = [[NSUserDefaults standardUserDefaults] stringForKey:kMessage1PreferenceKey];
    if (_wkStatus & 0x02) message = [[NSUserDefaults standardUserDefaults] stringForKey:kMessage2PreferenceKey];
    if (_wkStatus & 0x04) message = [[NSUserDefaults standardUserDefaults] stringForKey:kMessage3PreferenceKey];
    if (_wkStatus & 0x10) message = [[NSUserDefaults standardUserDefaults] stringForKey:kMessage4PreferenceKey];
    if (nil != message) {
        [self appendMessage:[self substituteTokens:message]];
    }
}

- (BOOL)setWK2Mode:(NSError**)error
{
    [self putToWinKeyer:bytesToData(kWKAdminSetWK2ModeCommand, 2)];
    return YES;
}

// MARK: - private serial comm methods -

- (void)hostOpen
{
    ORSSerialPacketDescriptor* descriptor = [[ORSSerialPacketDescriptor alloc] initWithMaximumPacketLength:1
                                                                                                  userInfo:nil
                                                                                         responseEvaluator:^BOOL(NSData* inputData) {
        if (inputData.length == 0) return NO;
        return YES;
    }];
    NSDictionary* userInfo = @{@"CommandName": @"HostOpen"};
    ORSSerialRequest* request = [ORSSerialRequest requestWithDataToSend:bytesToData(kWKAdminHostOpenCommand, 2)
                                                               userInfo:userInfo
                                                        timeoutInterval:10.0
                                                     responseDescriptor:descriptor];
    [winKeyerSerialPort sendRequest:request];
}

- (void)readMinorVersion
{
    ORSSerialPacketDescriptor* descriptor = [[ORSSerialPacketDescriptor alloc] initWithMaximumPacketLength:1
                                                                                                  userInfo:nil
                                                                                         responseEvaluator:^BOOL(NSData* inputData) {
        if (inputData.length == 0) return NO;
        return YES;
    }];
    NSDictionary* userInfo = @{@"CommandName": @"ReadMinorVersion"};
    ORSSerialRequest* request = [ORSSerialRequest requestWithDataToSend:bytesToData(kWKAdminReadMinorVersionCommand, 2)
                                                               userInfo:userInfo
                                                        timeoutInterval:10.0
                                                     responseDescriptor:descriptor];
    [winKeyerSerialPort sendRequest:request];
}

- (void)readType
{
    ORSSerialPacketDescriptor* descriptor = [[ORSSerialPacketDescriptor alloc] initWithMaximumPacketLength:1
                                                                                                  userInfo:nil
                                                                                         responseEvaluator:^BOOL(NSData* inputData) {
        if (inputData.length == 0) return NO;
        return YES;
    }];
    NSDictionary* userInfo = @{@"CommandName": @"ReadType"};
    ORSSerialRequest* request = [ORSSerialRequest requestWithDataToSend:bytesToData(kWKAdminGetTypeCommand, 2)
                                                               userInfo:userInfo
                                                        timeoutInterval:10.0
                                                     responseDescriptor:descriptor];
    [winKeyerSerialPort sendRequest:request];
}

- (void)echoTest
{
    NSDictionary* userInfo = @{@"CommandName": @"EchoTest"};

    ORSSerialPacketDescriptor* descriptor = [[ORSSerialPacketDescriptor alloc] initWithMaximumPacketLength:1
                                                                                                  userInfo:nil
                                                                                         responseEvaluator:^BOOL(NSData* inputData) {
        if (inputData.length != 1) return NO;
        uint8* echo = (uint8*)(inputData.bytes);
        if (echo[0] != kWKAdminEchoTestCommand[2]) return NO;
        return YES;
    }];
    ORSSerialRequest* request = [ORSSerialRequest requestWithDataToSend:bytesToData(kWKAdminEchoTestCommand, 3)
                                                               userInfo:userInfo
                                                        timeoutInterval:4.0
                                                     responseDescriptor:descriptor];
    [winKeyerSerialPort sendRequest:request];
}


- (BOOL)readVersion:(NSError**)error
{
    return YES;
}

- (BOOL)validateVersion:(NSError**)error
{
    NSUInteger version = [self.versionString integerValue];
    if (version < 20 || version > 23) {
        if (error) {
            NSString* errorDescription = [NSString stringWithFormat:@"WinKeyer returned unsupported version: %ld.", version];
            NSString* errorDetail = @"Confirm that the WinKeyer version is between 20 and 23.";
            NSDictionary* errorDict = @{NSLocalizedDescriptionKey: errorDescription, NSLocalizedRecoverySuggestionErrorKey: errorDetail};
            *error = [NSError errorWithDomain:@"me.K1GQ.SkookumMorse" code:1 userInfo:errorDict];
        }
        [[NSUserDefaults standardUserDefaults] setObject:kWKNoPortString forKey:kWKPortNamePreferenceKey];
        self.versionString = nil;
        [winKeyerSerialPort close];
        return NO;
    }
    return YES;
}

- (BOOL)determinePortName:(NSString**)portName error:(NSError**)error
{    
    NSString* aPortName = [[NSUserDefaults standardUserDefaults] stringForKey:kWKPortNamePreferenceKey];
    if ([aPortName isEqualToString:kWKNoPortString]) {
        if (error) {
            NSString* errorDescription = @"There is no port defined for the WinKeyer.";
            NSString* errorDetail = @"Choose a serial port other than None in the General Preferences pane. Or, plug in a cable to your WinKeyer USB.";
            NSDictionary* errorDict = @{NSLocalizedDescriptionKey: errorDescription, NSLocalizedRecoverySuggestionErrorKey: errorDetail};
            *error = [NSError errorWithDomain:@"me.K1GQ.SkookumMorse" code:1 userInfo:errorDict];
        }
        return NO;
    }
    *portName = aPortName;
    return YES;
}

- (BOOL)determineSerialPortForName:(NSString*)portName error:(NSError**)error
{
    for (ORSSerialPort* port in [ORSSerialPortManager sharedSerialPortManager].availablePorts) {
        //        NSLog(@"Serial port bsdPath: %@", [port bsdPath]);
        if ([portName compare:[port name]] == NSOrderedSame) {
            winKeyerSerialPort = port;
            return YES;
        }
    }
    if (error) {
        NSString* errorDescription = [NSString stringWithFormat:@"The serial port %@ is not available.", portName];
        NSString* errorDetail = @"Confirm that the WinKeyer is connected to the selected port, and that no other application is using that port.";
        NSDictionary* errorDict = @{NSLocalizedDescriptionKey: errorDescription, NSLocalizedRecoverySuggestionErrorKey: errorDetail};
        *error = [NSError errorWithDomain:@"me.K1GQ.SkookumMorse" code:1 userInfo:errorDict];
    }    
    [[NSUserDefaults standardUserDefaults] setObject:kWKNoPortString forKey:kWKPortNamePreferenceKey];
    winKeyerSerialPort = nil;
    return NO;
}

- (BOOL)openSerialPort:(NSError**)error
{
    [winKeyerSerialPort open];
    if (![winKeyerSerialPort isOpen]) {
        if (error) {
            NSString* errorDescription = [NSString stringWithFormat:@"Unable to open the WinKeyer serial port %@.", [winKeyerSerialPort name]];
            NSString* errorDetail = @"Confirm that no other app is using the port.";
            NSDictionary* errorDict = @{NSLocalizedDescriptionKey: errorDescription, NSLocalizedRecoverySuggestionErrorKey: errorDetail};
            *error = [NSError errorWithDomain:@"me.K1GQ.SkookumMorse" code:1 userInfo:errorDict];
        }
        [[NSUserDefaults standardUserDefaults] setObject:kWKNoPortString forKey:kWKPortNamePreferenceKey];
        [winKeyerSerialPort close];
        return NO;
    }
    return YES;
}

- (BOOL)configureSerialPort:(NSError**)error
{
    winKeyerSerialPort.baudRate = @(B1200);
    winKeyerSerialPort.parity = ORSSerialPortParityNone;
    winKeyerSerialPort.numberOfStopBits = 2L;
    // WK_USB Application Interface Guide v1.1 says enable DTR, but reads timeout if do so.
    winKeyerSerialPort.usesDTRDSRFlowControl = NO;
    winKeyerSerialPort.usesRTSCTSFlowControl = NO;
    winKeyerSerialPort.usesDCDOutputFlowControl = NO;
    return YES;
}

- (BOOL)syncWinKeyer:(NSError**)error
{
    static NSUInteger syncCount = 4;
    for (NSUInteger i = 0; i < syncCount; ++i) {
        [self putToWinKeyer:byteToData(kWKImmediateNullCommand)];
    }
    return YES;
}

@end
